import logging
from pathlib import Path
import socket
from adbutils import AdbConnection, AdbDevice, AdbError, Network
from fastapi import WebSocket
from retry import retry

logger = logging.getLogger(__name__)


class ScrcpyServer3:
    VERSION = "3.3.3"
    
    def __init__(self, device: AdbDevice):
        self._device = device
        self._shell_conn: AdbConnection
        self._video_sock: socket.socket
        self._control_sock: socket.socket
        
        self._shell_conn = self._start_scrcpy3()
        self._video_sock = self._connect_scrcpy(dummy_byte=True)
        self._control_sock = self._connect_scrcpy()
    
    def _start_scrcpy3(self):
        device = self._device
        jar_path = Path(__file__).parent.joinpath(f'../binaries/scrcpy-server-v{self.VERSION}.jar')
        device.sync.push(jar_path, '/data/local/tmp/scrcpy_server.jar', check=True)
        logger.info(f'{jar_path.name} pushed to device')

        # 构建启动 scrcpy 服务器的命令
        cmds = [
            'CLASSPATH=/data/local/tmp/scrcpy_server.jar',
            'app_process', '/',
            f'com.genymobile.scrcpy.Server', self.VERSION,
            'log_level=info', 'max_size=1024', 'max_fps=30',
            'video_bit_rate=8000000', 'tunnel_forward=true',
            'send_frame_meta=true',
            f'control=true',
            'audio=false', 'show_touches=false', 'stay_awake=false',
            'power_off_on_close=false', 'clipboard_autosync=false'
        ]
        conn = device.shell(cmds, stream=True)
        logger.debug("scrcpy output: %s", conn.conn.recv(100))
        return conn

    @retry(exceptions=AdbError, tries=20, delay=0.1)
    def _connect_scrcpy(self, dummy_byte: bool = False) -> socket.socket:
        sock = self._device.create_connection(Network.LOCAL_ABSTRACT, 'scrcpy')
        if dummy_byte:
            received = sock.recv(1)
            if not received or received != b"\x00":
                raise ConnectionError("Did not receive Dummy Byte!")
            logger.debug('Received Dummy Byte!')
        return sock
    
    def stream_to_websocket(self, ws: WebSocket):
        from .pipe import RWSocketDuplex, WebSocketDuplex, AsyncDuplex, pipe_duplex
        socket_duplex = RWSocketDuplex(self._video_sock, self._control_sock)
        websocket_duplex = WebSocketDuplex(ws)
        return pipe_duplex(socket_duplex, websocket_duplex)

    def close(self):
        self._safe_close_sock(self._control_sock)
        self._safe_close_sock(self._video_sock)
        self._shell_conn.close()
        
    def _safe_close_sock(self, sock: socket.socket):
        try:
            sock.close()
        except:
            pass
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json
import logging
import os
import re
import shutil
import subprocess
import tempfile
import time
import uuid
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union, final

from PIL import Image

from uiautodev.command_types import CurrentAppResponse
from uiautodev.driver.base_driver import BaseDriver
from uiautodev.model import AppInfo, Node, Rect, ShellResponse, WindowSize

logger = logging.getLogger(__name__)

StrOrPath = Union[str, Path]


def run_command(command: str, timeout: int = 60) -> str:
    try:
        result = subprocess.run(
            command,
            shell=True,
            capture_output=True,
            timeout=timeout,
            text=True,
            errors='ignore',
            input='' # this avoid stdout: "FreeChannelContinue handle->data is nullptr"
        )
        # the hdc shell stderr is (不仅没啥用，还没办法去掉)
        # Remote PTY will not be allocated because stdin is not a terminal.
        # Use multiple -t options to force remote PTY allocation.
        output = result.stdout.strip()
        return output
    except subprocess.TimeoutExpired as e:
        raise TimeoutError(f"{command:r} timeout {e}")


class HDCError(Exception):
    pass


class HDC:
    def __init__(self):
        self.hdc = 'hdc'
        self.tmpdir = tempfile.TemporaryDirectory()

    def __del__(self):
        self.tmpdir.cleanup()

    def list_device(self) -> List[str]:
        command = f"{self.hdc} list targets"
        result = run_command(command)
        if result and not "Empty" in result:
            devices = []
            for line in result.strip().split("\n"):
                serial = line.strip().split('\t', 1)[0]
                devices.append(serial)
            return devices
        else:
            return []

    def shell(self, serial: str, command: str) -> str:
        command = f"{self.hdc} -t {serial} shell \"{command}\""
        result = run_command(command)
        return result.strip()

    def __split_text(self, text: str) -> str:
        return text.split("\n")[0].strip() if text else ""

    def get_model(self, serial: str) -> str:
        return self.shell(serial, "param get const.product.model")

    def get_name(self, serial: str) -> str:
        data = self.shell(serial, "param get const.product.name")
        return self.__split_text(data)

    def wlan_ip(self, serial: str) -> str:
        data = self.shell(serial, "ifconfig")
        if not data or "not found" in data.lower() or "error" in data.lower():
            logger.warning(f"ifconfig command failed or returned error for serial {serial}: {data!r}")
            return ""
        # Try multiple patterns for IP address
        matches = re.findall(r'inet addr:(?!127)(\d+\.\d+\.\d+\.\d+)', data)
        if not matches:
            matches = re.findall(r'inet (?!127)(\d+\.\d+\.\d+\.\d+)', data)
        if matches:
            return matches[0]
        logger.warning(f"No valid IP address found in ifconfig output for serial {serial}: {data!r}")
        return ""

    def sdk_version(self, serial: str) -> str:
        data = self.shell(serial, "param get const.ohos.apiversion")
        return self.__split_text(data)

    def sys_version(self, serial: str) -> str:
        data = self.shell(serial, "param get const.product.software.version")
        return self.__split_text(data)

    def brand(self, serial: str) -> str:
        data = self.shell(serial, "param get const.product.brand")
        return self.__split_text(data)

    def pull(self, serial: str, remote: StrOrPath, local: StrOrPath):
        if isinstance(remote, Path):
            remote = remote.as_posix()
        command = f"{self.hdc} -t {serial} file recv {remote} {local}"
        output = run_command(command)
        if not os.path.exists(local):
            raise HDCError(f"device file: {remote} not found", output)

    def push(self, serial: str, local: StrOrPath, remote: StrOrPath) -> str:
        if isinstance(remote, Path):
            remote = remote.as_posix()
        command = f"{self.hdc} -t {serial} file send {local} {remote}"
        return run_command(command)

    def screenshot(self, serial: str) -> Image.Image:
        device_path = f'/data/local/tmp/screenshot-{int(time.time()*1000)}.png'
        self.shell(serial, f"uitest screenCap -p {device_path}")
        try:
            local_path = os.path.join(self.tmpdir.name, f"{uuid.uuid4()}.png")
            self.pull(serial, device_path, local_path)
            with Image.open(local_path) as image:
                image.load()
                return image
        finally:
            self.shell(serial, f"rm {device_path}")

    def dump_layout(self, serial: str) -> dict:
        name = "{}.json".format(int(time.time() * 1000))
        remote_path = f"/data/local/tmp/layout-{name}.json"
        temp_path = os.path.join(self.tmpdir.name, f"layout-{name}.json")
        output = self.shell(serial, f"uitest dumpLayout -p {remote_path}")
        self.pull(serial, remote_path, temp_path)
        # mock
        # temp_path = Path(__file__).parent / 'testdata/layout.json'
        try:
            with open(temp_path, "rb") as f:
                json_content = json.load(f)
            return json_content
        except json.JSONDecodeError:
            raise HDCError(f"failed to dump layout: {output}")
        finally:
            self.shell(serial, f"rm {remote_path}")


class HarmonyDriver(BaseDriver):
    def __init__(self, hdc: HDC, serial: str):
        super().__init__(serial)
        self.hdc = hdc

    def screenshot(self, id: int = 0) -> Image.Image:
        return self.hdc.screenshot(self.serial)

    def window_size(self) -> WindowSize:
        result = self.hdc.shell(self.serial, "hidumper -s 10 -a screen")
        pattern = r"activeMode:\s*(\d+x\d+)"
        match = re.search(pattern, result)
        if match:
            resolution = match.group(1).split("x")
            return WindowSize(width=int(resolution[0]), height=int(resolution[1]))
        else:
            image = self.screenshot()
            return WindowSize(width=image.width, height=image.height)

    def dump_hierarchy(self) -> Tuple[str, Node]:
        """returns xml string and hierarchy object"""
        layout = self.hdc.dump_layout(self.serial)
        return json.dumps(layout), parse_json_element(layout)

    def tap(self, x: int, y: int):
        self.hdc.shell(self.serial, f"uinput -T -c {x} {y}")

    def app_current(self) -> Optional[CurrentAppResponse]:
        echo = self.hdc.shell(self.serial, "hidumper -s WindowManagerService -a '-a'")
        focus_window = re.search(r"Focus window: (\d+)", echo)
        if focus_window:
            focus_window = focus_window.group(1)
        mission_echo = self.hdc.shell(self.serial, "aa dump -a")
        pkg_names = re.findall(r"Mission ID #(\d+)\s+mission name #\[(.*?)\]", mission_echo)
        if focus_window and pkg_names:
            for mission in pkg_names:
                mission_id = mission[0]
                if focus_window == mission_id:
                    mission_name = mission[1]
                    pkg_name = mission_name.split(":")[0].replace("#", "")
                    ability_name = mission_name.split(":")[-1]
                    pid = self.hdc.shell(self.serial, f"pidof {pkg_name}").strip()
                    return CurrentAppResponse(package=pkg_name, activity=ability_name, pid=int(pid))
        else:
            return None

    def get_app_info(self, package_name: str) -> Dict:
        """
        Get detailed information about a specific application.

        Args:
            package_name (str): The package name of the application to retrieve information for.

        Returns:
            Dict: A dictionary containing the application information. If an error occurs during parsing,
                  an empty dictionary is returned.
        """
        app_info = {}
        data = self.hdc.shell(self.serial, f"bm dump -n {package_name}")
        output = data
        try:
            json_start = output.find("{")
            json_end = output.rfind("}") + 1
            json_output = output[json_start:json_end]

            app_info = json.loads(json_output)
        except Exception as e:
            logger.error(f"An error occurred: {e}")
        return app_info

    def get_app_abilities(self, package_name: str) -> List[Dict]:
        """
        Get the abilities of an application.

        Args:
            package_name (str): The package name of the application.

        Returns:
            List[Dict]: A list of dictionaries containing the abilities of the application.
        """
        result = []
        app_info = self.get_app_info(package_name)
        hap_module_infos = app_info.get("hapModuleInfos")
        main_entry = app_info.get("mainEntry")
        for hap_module_info in hap_module_infos:
            # 尝试读取moduleInfo
            try:
                ability_infos = hap_module_info.get("abilityInfos")
                module_main = hap_module_info["mainAbility"]
            except Exception as e:
                logger.warning(f"Fail to parse moduleInfo item, {repr(e)}")
                continue
            # 尝试读取abilityInfo
            for ability_info in ability_infos:
                try:
                    is_launcher_ability = False
                    skills = ability_info['skills']
                    if len(skills) > 0 and "action.system.home" in skills[0]["actions"]:
                        is_launcher_ability = True
                    icon_ability_info = {
                        "name": ability_info["name"],
                        "moduleName": ability_info["moduleName"],
                        "moduleMainAbility": module_main,
                        "mainModule": main_entry,
                        "isLauncherAbility": is_launcher_ability
                    }
                    result.append(icon_ability_info)
                except Exception as e:
                    logger.warning(f"Fail to parse ability_info item, {repr(e)}")
                    continue
        logger.debug(f"all abilities: {result}")
        return result

    def get_app_main_ability(self, package_name: str) -> Dict:
        """
        Get the main ability of an application.

        Args:
            package_name (str): The package name of the application to retrieve information for.

        Returns:
            Dict: A dictionary containing the main ability of the application.

        """
        if not (abilities := self.get_app_abilities(package_name)):
            return {}
        for item in abilities:
            score = 0
            if (name := item["name"]) and name == item["moduleMainAbility"]:
                score += 1
            if (module_name := item["moduleName"]) and module_name == item["mainModule"]:
                score += 1
            item["score"] = score
        abilities.sort(key=lambda x: (not x["isLauncherAbility"], -x["score"]))
        logger.debug(f"main ability: {abilities[0]}")
        return abilities[0]

    def app_launch(self, package: str, page_name: Optional[str] = None):
        """
        Start an application on the device.
        If the `page_name` is empty, it will retrieve the main ability using `get_app_main_ability`.
        Args:
            package (str): The package name of the application.
            page_name (Optional[str]): Ability Name within the application to start. If not provided, the main ability will be used.
        """
        if not page_name:
            page_name = self.get_app_main_ability(package).get('name', 'MainAbility')
        self.shell(f"aa start -a {page_name} -b {package}")

    def app_terminate(self, package: str):
        self.shell(f"aa force-stop {package}")

    def shell(self, command: str) -> ShellResponse:
        result = self.hdc.shell(self.serial, command)
        return ShellResponse(output=result)

    def home(self):
        self.hdc.shell(self.serial, "uinput -K -d 1 -u 1")

    def back(self):
        self.hdc.shell(self.serial, "uinput -K -d 2 -u 2")

    def volume_up(self):
        self.hdc.shell(self.serial, "uinput -K -d 16 -u 16")

    def volume_down(self):
        self.hdc.shell(self.serial, "uinput -K -d 17 -u 17")

    def volume_mute(self):
        self.hdc.shell(self.serial, "uinput -K -d 22 -u 22")

    def app_switch(self):
        self.hdc.shell(self.serial, "uinput -K -d 2076 -d 2049 -u 2076 -u 2049")

    def app_list(self) -> List[AppInfo]:
        results = []
        output = self.hdc.shell(self.serial, "bm dump -a")
        for i in output.split("\n"):
            if "ID" in i:
                continue
            else:
                results.append(AppInfo(packageName=i.strip()))
        return results


def parse_json_element(element, indexes: List[int] = [0]) -> Node:
    """
    Recursively parse an json element into a dictionary format.
    """
    attributes = element.get("attributes", {})
    name = attributes.get("type", "")
    bounds = attributes.get("bounds", "")
    bounds = list(map(int, re.findall(r"\d+", bounds)))
    assert len(bounds) == 4
    rect = Rect(x=bounds[0], y=bounds[1], width=bounds[2] - bounds[0], height=bounds[3] - bounds[1])
    elem = Node(
        key="-".join(map(str, indexes)),
        name=name,
        bounds=None,
        rect=rect,
        properties={key: attributes[key] for key in attributes},
        children=[],
    )
    # Construct xpath for children
    for index, child in enumerate(element.get("children", [])):
        child_node = parse_json_element(child, indexes + [index])
        if child_node:
            elem.children.append(child_node)

    return elem



import asyncio
import json
import logging
import socket
from datetime import datetime
from threading import Thread

from fastapi import WebSocket
from hypium import KeyCode

from uiautodev.exceptions import HarmonyDriverException

logger = logging.getLogger(__name__)


class HarmonyMjpegServer:
    """
    HarmonyMjpegServer is responsible for handling screen streaming functionality
    for HarmonyOS devices that support ABC proxy (a communication interface).

    It manages WebSocket clients, communicates with the ABC server over gRPC, and streams
    the device's screen data in real-time to connected clients.

    This server is specifically designed for devices running in 'abc mode' and requires that
    the target device expose an `abc_proxy` attribute for communication.

    Attributes:
        device: The HarmonyOS device object.
        driver: The controlling driver which may wrap the device.
        abc_rpc_addr: Tuple containing the IP and port used to communicate with abc_proxy.
        channel: The gRPC communication channel (initialized later).
        clients: A set of connected WebSocket clients.
        loop: Asyncio event loop used to run asynchronous tasks.
        is_running: Boolean flag indicating if the streaming service is active.

    Raises:
        RuntimeError: If the connected device does not support abc_proxy.

    References:
        - Huawei HarmonyOS Python Guidelines:
          https://developer.huawei.com/consumer/cn/doc/harmonyos-guides/hypium-python-guidelines
    """

    def __init__(self, driver):
        if hasattr(driver, "_device"):
            device = driver._device
        else:
            device = driver
        logger.info(f'device: {device}')
        if not hasattr(device, "abc_proxy") or device.abc_proxy is None:
            raise HarmonyDriverException("Only abc mode can support screen recorder")
        self.device = device
        self.driver = driver
        self.abc_rpc_addr = ("127.0.0.1", device.abc_proxy.port)
        self.channel = None
        self.clients = set()
        self.loop = asyncio.get_event_loop()
        self.is_running = False

    def connect(self):
        self.channel = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.channel.connect(self.abc_rpc_addr)

    def start(self, timeout=3600):
        if self.channel is None:
            self.connect()
        self.is_running = True
        self.timeout = timeout
        self.stop_capture_if_running()
        msg_json = {'api': "startCaptureScreen", 'args': []}
        full_msg = {
            "module": "com.ohos.devicetest.hypiumApiHelper",
            "method": "Captures", 
            "params": msg_json,
            "request_id": datetime.now().strftime("%Y%m%d%H%M%S%f")
        }
        full_msg_str = json.dumps(full_msg, ensure_ascii=False, separators=(',', ':'))
        self.channel.sendall(full_msg_str.encode("utf-8") + b'\n')
        reply = self.channel.recv(1024)
        logger.info(f'reply: {reply}')
        if b"true" in reply:
            thread_record = Thread(target=self._record_worker)
            thread_record.start()
        else:
            raise RuntimeError("Fail to start screen capture")

    def stop_capture_if_running(self):
        msg_json = {'api': "stopCaptureScreen", 'args': []}
        full_msg = {
            "module": "com.ohos.devicetest.hypiumApiHelper",
            "method": "Captures", 
            "params": msg_json,
            "request_id": datetime.now().strftime("%Y%m%d%H%M%S%f")
        }
        full_msg_str = json.dumps(full_msg, ensure_ascii=False, separators=(',', ':'))
        self.channel.sendall(full_msg_str.encode("utf-8") + b'\n')
        reply = self.channel.recv(1024)
        logger.info(f'stop reply: {reply}')

    async def handle_ws(self, websocket: WebSocket):
        self.clients.add(websocket)
        serial = getattr(self.device, "device_sn", "unknown")
        logger.info(f"[{serial}] WebSocket connected")

        try:
            while True:
                message = await websocket.receive_text()
                logger.info(f"Received message: {message}")
                try:
                    data = json.loads(message)
                    if data.get('type') == 'touch':
                        action = data.get('action')
                        x, y = data.get('x'), data.get('y')
                        if action == 'normal':
                            self.driver.touch((x, y))
                        elif action == 'long':
                            self.driver.touch(target=(x, y), mode='long')
                        elif action == 'double':
                            self.driver.touch(target=(x, y), mode='double')
                        elif action == 'move':
                            self.driver.slide(
                                start=(data.get('x1'), data.get('y1')),
                                end=(data.get('x2'), data.get('y2')),
                                slide_time=0.1
                            )
                    elif data.get('type') == 'keyEvent':
                        event_number = data['eventNumber']
                        if event_number == 187:
                            self.driver.swipe_to_recent_task()
                        elif event_number == 3:
                            self.driver.go_home()
                        elif event_number == 4:
                            self.driver.go_back()
                        elif event_number == 224:
                            self.driver.wake_up_display()
                    elif data.get('type') == 'text':
                        detail = data.get('detail')
                        if detail == 'CODE_AC_BACK':
                            self.driver.press_key(KeyCode.DEL)
                        elif detail == 'CODE_AC_ENTER':
                            self.driver.press_key(KeyCode.ENTER)
                        else:
                            self.driver.shell(
                                f"uitest uiInput inputText {data.get('x')} {data.get('y')} {detail}")
                except Exception as e:
                    logger.warning(f"Failed to handle message: {e}")
        except Exception as e:
            logger.info(f"WebSocket closed: {e}")
        finally:
            self.clients.discard(websocket)

    def _record_worker(self):
        tmp_data = b''
        start_flag = b'\xff\xd8'
        end_flag = b'\xff\xd9'
        while self.is_running:
            try:
                result = self.channel.recv(4096 * 1024)
                tmp_data += result
                while start_flag in tmp_data and end_flag in tmp_data:
                    start_index = tmp_data.index(start_flag)
                    end_index = tmp_data.index(end_flag) + 2
                    frame = tmp_data[start_index:end_index]
                    tmp_data = tmp_data[end_index:]
                    asyncio.run_coroutine_threadsafe(self._broadcast(frame), self.loop)
            except Exception as e:
                logger.warning(f"Record worker error: {e}")
                self.is_running = False
                self.channel = None
                break

    async def _broadcast(self, data):
        for client in self.clients.copy():
            try:
                await client.send_bytes(data)
            except Exception as e:
                logger.info(f"Send error, removing client: {e}")
                self.clients.discard(client)

    def stop(self):
        self.is_running = False
        if self.channel is None:
            return
        msg_json = {'api': "stopCaptureScreen", 'args': []}
        full_msg = {
            "module": "com.ohos.devicetest.hypiumApiHelper",
            "method": "Captures",
            "params": msg_json,
            "request_id": datetime.now().strftime("%Y%m%d%H%M%S%f")
        }
        full_msg_str = json.dumps(full_msg, ensure_ascii=False, separators=(',', ':'))
        self.channel.sendall(full_msg_str.encode("utf-8") + b'\n')
        reply = self.channel.recv(1024)
        if b"true" not in reply:
            logger.info("Fail to stop capture")
        self.channel.close()
        self.channel = None
        for client in self.clients:
            asyncio.run_coroutine_threadsafe(client.close(), self.loop)



import asyncio
import socket
from typing import Optional, Protocol
from starlette.websockets import WebSocket, WebSocketDisconnect


class AsyncDuplex(Protocol):
    async def read(self, n: int = -1) -> bytes: ...
    async def write(self, data: bytes): ...
    async def close(self): ...
    

async def pipe_duplex(a: AsyncDuplex, b: AsyncDuplex, label_a="A", label_b="B"):
    """双向管道：a <-> b"""
    task_ab = asyncio.create_task(_pipe_oneway(a, b, f"{label_a}->{label_b}"))
    task_ba = asyncio.create_task(_pipe_oneway(b, a, f"{label_b}->{label_a}"))
    done, pending = await asyncio.wait(
        [task_ab, task_ba],
        return_when=asyncio.FIRST_COMPLETED,
    )
    for t in pending:
        t.cancel()
    if pending:
        await asyncio.gather(*pending, return_exceptions=True)


async def _pipe_oneway(src: AsyncDuplex, dst: AsyncDuplex, name: str):
    try:
        while True:
            data = await src.read(4096)
            if not data:
                break
            await dst.write(data)
    except asyncio.CancelledError:
        pass
    except Exception as e:
        print(f"[{name}] error:", e)
    finally:
        await dst.close()
        
class RWSocketDuplex:
    def __init__(self, rsock: socket.socket, wsock: socket.socket, loop=None):
        self.rsock = rsock
        self.wsock = wsock
        self._same = rsock is wsock
        self.loop = loop or asyncio.get_running_loop()
        self._closed = False

        self.rsock.setblocking(False)
        if not self._same:
            self.wsock.setblocking(False)

    async def read(self, n: int = 4096) -> bytes:
        if self._closed:
            return b''
        try:
            data = await self.loop.sock_recv(self.rsock, n)
            if not data:
                await self.close()
                return b''
            return data
        except (ConnectionResetError, OSError):
            await self.close()
            return b''
        
    async def write(self, data: bytes):
        if not data or self._closed:
            return
        try:
            await self.loop.sock_sendall(self.wsock, data)
        except (ConnectionResetError, OSError):
            await self.close()

    async def close(self):
        if self._closed:
            return
        self._closed = True
        try:
            self.rsock.close()
        except Exception:
            pass
        if not self._same:
            try:
                self.wsock.close()
            except Exception:
                pass

    def is_closed(self):
        return self._closed
    
class SocketDuplex(RWSocketDuplex):
    """封装 socket.socket 为 AsyncDuplex 接口"""
    def __init__(self, sock: socket.socket, loop: Optional[asyncio.AbstractEventLoop] = None):
        super().__init__(sock, sock, loop)


class WebSocketDuplex:
    """将 starlette.websockets.WebSocket 封装为 AsyncDuplex"""
    def __init__(self, ws: WebSocket):
        self.ws = ws
        self._closed = False

    async def read(self, n: int = -1) -> bytes:
        """读取二进制消息，如果是文本则自动转 bytes"""
        if self._closed:
            return b''
        try:
            msg = await self.ws.receive()
        except WebSocketDisconnect:
            self._closed = True
            return b''
        except Exception:
            self._closed = True
            return b''

        if msg["type"] == "websocket.disconnect":
            self._closed = True
            return b''
        elif msg["type"] == "websocket.receive":
            data = msg.get("bytes")
            if data is not None:
                return data
            text = msg.get("text")
            return text.encode("utf-8") if text else b''
        return b''

    async def write(self, data: bytes):
        if self._closed:
            return
        try:
            await self.ws.send_bytes(data)
        except Exception:
            self._closed = True

    async def close(self):
        if not self._closed:
            self._closed = True
            try:
                await self.ws.close()
            except Exception:
                pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Created on Fri Mar 01 2024 14:19:29 by codeskyblue
"""

import logging
import re
import time
from functools import cached_property
from typing import Optional, Tuple

import uiautomator2 as u2
from PIL import Image

from uiautodev.driver.android.adb_driver import ADBAndroidDriver
from uiautodev.driver.android.common import parse_xml
from uiautodev.exceptions import AndroidDriverException
from uiautodev.model import AppInfo, Node, WindowSize

logger = logging.getLogger(__name__)

class U2AndroidDriver(ADBAndroidDriver):
    def __init__(self, serial: str):
        super().__init__(serial)

    @cached_property
    def ud(self) -> u2.Device:
        return u2.connect_usb(self.serial)
    
    def screenshot(self, id: int) -> Image.Image:
        if id > 0:
            # u2 is not support multi-display yet
            return super().screenshot(id)
        return self.ud.screenshot()

    def dump_hierarchy(self, display_id: Optional[int] = 0) -> Tuple[str, Node]:
        """returns xml string and hierarchy object"""
        start = time.time()
        xml_data = self._dump_hierarchy_raw()
        logger.debug("dump_hierarchy cost: %s", time.time() - start)

        wsize = self.adb_device.window_size()
        logger.debug("window size: %s", wsize)
        return xml_data, parse_xml(
            xml_data, WindowSize(width=wsize[0], height=wsize[1]), display_id
        )

    def _dump_hierarchy_raw(self) -> str:
        """
        uiautomator2 server is conflict with "uiautomator dump" command.

        uiautomator dump errors:
        - ERROR: could not get idle state.
        """
        try:
            return self.ud.dump_hierarchy()
        except Exception as e:
            raise AndroidDriverException(f"Failed to dump hierarchy: {str(e)}")
    
    def tap(self, x: int, y: int):
        self.ud.click(x, y)
    
    def send_keys(self, text: str):
        self.ud.send_keys(text)
    
    def clear_text(self):
        self.ud.clear_text()